#version 450
#extension GL_ARB_separate_shader_objects : enable

layout(location = 0) in VS_OUT {
    vec2 texCoords;
} vs_out;

layout(push_constant) uniform Uniforms {
    int index;
    float seed;
    float resolution;
    float res1;
    float res2;
    float resMix;
    float mixScale;
    float doesRidged;
} uniforms;

const int octaves = 16;

layout(location = 0) out vec4 outColor;

// ---------------------------------------------------------------------------------------------------------------------
// GLSL textureless classic 4D noise "cnoise",
// with an RSL-style periodic variant "pnoise".
// Author:  Stefan Gustavson (stefan.gustavson@liu.se)
// Version: 2011-08-22
//
// Many thanks to Ian McEwan of Ashima Arts for the
// ideas for permutation and gradient selection.
//
// Copyright (c) 2011 Stefan Gustavson. All rights reserved.
// Distributed under the MIT license. See LICENSE file.
// https://github.com/ashima/webgl-noise
//

//
// GLSL textureless classic 3D noise "cnoise",
// with an RSL-style periodic variant "pnoise".
// Author:  Stefan Gustavson (stefan.gustavson@liu.se)
// Version: 2011-10-11
//
// Many thanks to Ian McEwan of Ashima Arts for the
// ideas for permutation and gradient selection.
//
// Copyright (c) 2011 Stefan Gustavson. All rights reserved.
// Distributed under the MIT license. See LICENSE file.
// https://github.com/ashima/webgl-noise
//

vec3 mod289(vec3 x)
{
  return x - floor(x * (1.0 / 289.0)) * 289.0;
}

vec4 mod289(vec4 x)
{
  return x - floor(x * (1.0 / 289.0)) * 289.0;
}

vec4 permute(vec4 x)
{
  return mod289(((x*34.0)+1.0)*x);
}

vec4 taylorInvSqrt(vec4 r)
{
  return 1.79284291400159 - 0.85373472095314 * r;
}

vec3 fade(vec3 t) {
  return t*t*t*(t*(t*6.0-15.0)+10.0);
}

// Classic Perlin noise
float cnoise(vec3 P)
{
  vec3 Pi0 = floor(P); // Integer part for indexing
  vec3 Pi1 = Pi0 + vec3(1.0); // Integer part + 1
  Pi0 = mod289(Pi0);
  Pi1 = mod289(Pi1);
  vec3 Pf0 = fract(P); // Fractional part for interpolation
  vec3 Pf1 = Pf0 - vec3(1.0); // Fractional part - 1.0
  vec4 ix = vec4(Pi0.x, Pi1.x, Pi0.x, Pi1.x);
  vec4 iy = vec4(Pi0.yy, Pi1.yy);
  vec4 iz0 = Pi0.zzzz;
  vec4 iz1 = Pi1.zzzz;

  vec4 ixy = permute(permute(ix) + iy);
  vec4 ixy0 = permute(ixy + iz0);
  vec4 ixy1 = permute(ixy + iz1);

  vec4 gx0 = ixy0 * (1.0 / 7.0);
  vec4 gy0 = fract(floor(gx0) * (1.0 / 7.0)) - 0.5;
  gx0 = fract(gx0);
  vec4 gz0 = vec4(0.5) - abs(gx0) - abs(gy0);
  vec4 sz0 = step(gz0, vec4(0.0));
  gx0 -= sz0 * (step(0.0, gx0) - 0.5);
  gy0 -= sz0 * (step(0.0, gy0) - 0.5);

  vec4 gx1 = ixy1 * (1.0 / 7.0);
  vec4 gy1 = fract(floor(gx1) * (1.0 / 7.0)) - 0.5;
  gx1 = fract(gx1);
  vec4 gz1 = vec4(0.5) - abs(gx1) - abs(gy1);
  vec4 sz1 = step(gz1, vec4(0.0));
  gx1 -= sz1 * (step(0.0, gx1) - 0.5);
  gy1 -= sz1 * (step(0.0, gy1) - 0.5);

  vec3 g000 = vec3(gx0.x,gy0.x,gz0.x);
  vec3 g100 = vec3(gx0.y,gy0.y,gz0.y);
  vec3 g010 = vec3(gx0.z,gy0.z,gz0.z);
  vec3 g110 = vec3(gx0.w,gy0.w,gz0.w);
  vec3 g001 = vec3(gx1.x,gy1.x,gz1.x);
  vec3 g101 = vec3(gx1.y,gy1.y,gz1.y);
  vec3 g011 = vec3(gx1.z,gy1.z,gz1.z);
  vec3 g111 = vec3(gx1.w,gy1.w,gz1.w);

  vec4 norm0 = taylorInvSqrt(vec4(dot(g000, g000), dot(g010, g010), dot(g100, g100), dot(g110, g110)));
  g000 *= norm0.x;
  g010 *= norm0.y;
  g100 *= norm0.z;
  g110 *= norm0.w;
  vec4 norm1 = taylorInvSqrt(vec4(dot(g001, g001), dot(g011, g011), dot(g101, g101), dot(g111, g111)));
  g001 *= norm1.x;
  g011 *= norm1.y;
  g101 *= norm1.z;
  g111 *= norm1.w;

  float n000 = dot(g000, Pf0);
  float n100 = dot(g100, vec3(Pf1.x, Pf0.yz));
  float n010 = dot(g010, vec3(Pf0.x, Pf1.y, Pf0.z));
  float n110 = dot(g110, vec3(Pf1.xy, Pf0.z));
  float n001 = dot(g001, vec3(Pf0.xy, Pf1.z));
  float n101 = dot(g101, vec3(Pf1.x, Pf0.y, Pf1.z));
  float n011 = dot(g011, vec3(Pf0.x, Pf1.yz));
  float n111 = dot(g111, Pf1);

  vec3 fade_xyz = fade(Pf0);
  vec4 n_z = mix(vec4(n000, n100, n010, n110), vec4(n001, n101, n011, n111), fade_xyz.z);
  vec2 n_yz = mix(n_z.xy, n_z.zw, fade_xyz.y);
  float n_xyz = mix(n_yz.x, n_yz.y, fade_xyz.x);
  return 2.2 * n_xyz;
}
// ---------------------------------------------------------------------------------------------------------------------

// Source: https://github.com/colordodge/ProceduralPlanet by colordodge
vec3 getSphericalCoord(int index, float x, float y, float width) {
  width /= 2.0;
  x -= width;
  y -= width;
  vec3 coord = vec3(0.0, 0.0, 0.0);

  if (index == 0) {coord.x=width; coord.y=-y; coord.z=-x;}
  else if (index == 1) {coord.x=-width; coord.y=-y; coord.z=x;}
  else if (index == 2) {coord.x=x; coord.y=width; coord.z=y;}
  else if (index == 3) {coord.x=x; coord.y=-width; coord.z=-y;}
  else if (index == 4) {coord.x=x; coord.y=-y; coord.z=width;}
  else if (index == 5) {coord.x=-x; coord.y=-y; coord.z=-width;}

  return normalize(coord);
}


float simplexRidged(vec3 pos, float seed) {
  float n = cnoise(vec3(pos + seed));
  n = (n + 1.0) * 0.5;
  n = 2.0 * (0.5 - abs(0.5 - n));
  return n;
}

float simplex(vec3 pos, float seed) {
  float n = cnoise(vec3(pos + seed));
  return (n + 1.0) * 0.5;
}

float baseNoise(vec3 pos, float frq, float seed ) {
  float amp = 0.5;

  float n = 0.0;
  float gain = 1.0;
  for(int i=0; i<octaves; i++) {
    n +=  simplex(vec3(pos.x*gain/frq, pos.y*gain/frq, pos.z*gain/frq), seed+float(i)*10.0) * amp/gain;
    gain *= 2.0;
  }

  // increase contrast
  n = ( (n - 0.5) * 2.0 ) + 0.5;

  return n;
}

float ridgedNoise(vec3 pos, float frq, float seed) {
  float amp = 0.5;
  float n = 0.0;
  float gain = 1.0;
  for(int i=0; i<octaves; i++) {
    n +=  simplexRidged(vec3(pos.x*gain/frq, pos.y*gain/frq, pos.z*gain/frq), seed+float(i)*10.0) * amp/gain;
    gain *= 2.0;
  }

  n = pow(n, 4.0);
  return n;
}

float invRidgedNoise(vec3 pos, float frq, float seed) {

  float amp = 0.5;

  float n = 0.0;
  float gain = 1.0;
  for(int i=0; i<octaves; i++) {
    n +=  simplexRidged(vec3(pos.x*gain/frq, pos.y*gain/frq, pos.z*gain/frq), seed+float(i)*10.0) * amp/gain;
    gain *= 2.0;
  }

  n = pow(n, 4.0);
  n = 1.0 - n;

  return n;
}

float cloud(vec3 pos, float seed) {
  float n = cnoise(vec3(pos + seed));
  // n = sin(n*4.0 * cos(n*2.0));
  n = sin(n*5.0);
  // n = abs(sin(n*5.0));
  // n = 1.0 - n;

  n = n*0.5 + 0.5;
  // n = 1.0-n;
  // n = n*1.2;
  // n = 1.0-n;

  return n;
}

float cloudNoise(vec3 pos, float frq, float seed) {

  float amp = 0.5;

  float n = 0.0;
  float gain = 1.0;
  for(int i=0; i<octaves; i++) {
    n +=  cloud(vec3(pos.x*gain/frq, pos.y*gain/frq, pos.z*gain/frq), seed+float(i)*10.0) * amp/gain;
    gain *= 2.0;
  }

  // n = pow(n, 5.0);

  n = 1.0-n;
  n = pow(n, 1.0);
  n = 1.0-n;

  return n;
}

void main() {
  float x = vs_out.texCoords.x;
  float y = vs_out.texCoords.y;
  vec3 sphericalCoord = getSphericalCoord(uniforms.index, x*uniforms.resolution, y*uniforms.resolution, uniforms.resolution);

  float sub1, sub2, sub3, n;

  float resMod = 1.0; // overall res magnification
  float resMod2 = uniforms.mixScale; // minimum res mod

  if (uniforms.doesRidged == 0.0) {
    sub1 = cloudNoise(sphericalCoord, uniforms.res1*resMod, uniforms.seed+11.437);
    sub2 = cloudNoise(sphericalCoord, uniforms.res2*resMod, uniforms.seed+93.483);
    sub3 = cloudNoise(sphericalCoord, uniforms.resMix*resMod, uniforms.seed+23.675);
    n = cloudNoise(sphericalCoord + vec3((sub1/sub3)*0.1), resMod2+sub2, uniforms.seed+78.236);
  }
  else if (uniforms.doesRidged == 1.0) {
    sub1 = ridgedNoise(sphericalCoord, uniforms.res1*resMod, uniforms.seed+83.706);
    sub2 = ridgedNoise(sphericalCoord, uniforms.res2*resMod, uniforms.seed+29.358);
    sub3 = ridgedNoise(sphericalCoord, uniforms.resMix*resMod, uniforms.seed+53.041);
    n = ridgedNoise(sphericalCoord + vec3((sub1/sub3)*0.1), resMod2+sub2, uniforms.seed+34.982);
  }
  else  if (uniforms.doesRidged == 2.0) {
    sub1 = invRidgedNoise(sphericalCoord, uniforms.res1*resMod, uniforms.seed+49.684);
    sub2 = invRidgedNoise(sphericalCoord, uniforms.res2*resMod, uniforms.seed+136.276);
    sub3 = invRidgedNoise(sphericalCoord, uniforms.resMix*resMod, uniforms.seed+3.587);
    n = invRidgedNoise(sphericalCoord + vec3((sub1/sub3)*0.1), resMod2+sub2, uniforms.seed+33.321);
  }
  else {
    sub1 = baseNoise(sphericalCoord, uniforms.res1*resMod, uniforms.seed+52.284);
    sub2 = baseNoise(sphericalCoord, uniforms.res2*resMod, uniforms.seed+137.863);
    sub3 = baseNoise(sphericalCoord, uniforms.resMix*resMod, uniforms.seed+37.241);
    float alpha = sub1*3.14*2.0;
    float beta = sub2*3.14*2.0;
    float fx = cos(alpha)*cos(beta);
    float fz = sin(alpha)*cos(beta);
    float fy = sin(beta);
    n = baseNoise(sphericalCoord + (vec3(fx,fy,fz) * sub3), 1.0, uniforms.seed+28.634);
  }

  outColor = vec4(pow(vec3(n), vec3(1.5)), 1.0);
}
